#!/usr/bin/env python3

"""
Parenthesized prefix notation lexer & parser
"""

from __future__ import annotations

import sys
import enum
import typing
import logging
import argparse
import unicodedata


LOGGER = logging.getLogger(__name__)


### ----------------------- error handling utilities ----------------------- ###

def read_unicode_scalars_of_text_file(filepath: str, input_encoding='utf-8') -> typing.Tuple[int, str]:
    """Read the content of a text file from `filepath` assuming text encoding `input_encoding`
    and provide a generator for tuples (I, S) where I is the byte offset as integer
    and S is the Unicode scalar as string of length 1.
    
    The concatenation of all Unicode scalars does *not* resemble the original file,
    because universal newlines are implemented. Hence, the sequences [CARRIAGE RETURN]
    or [CARRIAGE RETURN, LINE FEED] are returned as [LINE FEED] with the position
    of the first character.
    """
    with open(filepath, 'rt', encoding=input_encoding, newline='') as fd:
        byte_offset = 0
        was_carriage_return = False

        # iterate as long as we find Unicode scalars
        while True:
            unicode_scalar = fd.read(1)
            if not unicode_scalar:
                break

            # case distinction to support newline normalization
            if was_carriage_return and unicode_scalar == '\n':
                yield (byte_offset - 1, '\n')
            elif not was_carriage_return and unicode_scalar == '\r':
                pass
            else:
                if was_carriage_return:
                    yield (byte_offset - 1, '\n')
                yield (byte_offset, unicode_scalar)
            
            # prepare for next iteration
            byte_offset += len(unicode_scalar.encode(input_encoding))
            was_carriage_return = unicode_scalar == '\r'

        # was_carriage_return=True indicates that the last new line was not emitted
        if was_carriage_return:
            yield (byte_offset - 1, '\n')


def byte_range_to_position_information(filepath: str, byte_range: typing.Tuple[int, int], input_encoding='utf-8') -> typing.Tuple[typing.Tuple[int, int, int], typing.Tuple[int, int, int], str]:
    """Read text file `filepath` with text encoding `input_encoding` and look for the
    content between ``byte_range[0]`` and ``byte_range[1]`` as byte offsets respectively.
    Return the tuple (B, E, S) where B and E are tuples with (byte offset, line index,
    column index) as begin and end of the range respectively. S is the substring resulting
    from `byte_range` but with normalized new lines. Hence, the sequences [CARRIAGE RETURN]
    or [CARRIAGE RETURN, LINE FEED] are returned as [LINE FEED].
    
    NOTE: index means here that lines and columns are counted zero-based.
    NOTE: possible line breaks are '\n', '\r', '\r\n' in accordance with
          the python's newline parameter of the io library
    """
    # initialize algorithmic variables
    was_carriage_return = False
    line_index = 0
    column_index = 0
    byte_offset = 0

    # initialize result variables
    begin, end = None, None
    substring = ''  # TODO why is substring a string and not bytes?

    with open(filepath, 'rt', encoding=input_encoding, newline='') as fd:
        # iterate as long as we find Unicode scalars
        while True:
            unicode_scalar = fd.read(1)
            if not unicode_scalar:
                break

            # are we within the given byte_range?
            if byte_range[0] <= byte_offset <= byte_range[1]:
                substring += unicode_scalar
                if begin is None:
                    begin = (byte_offset, line_index, column_index)
                else:
                    end = (byte_offset, line_index, column_index)

            # case distinction to support newline normalization
            if unicode_scalar == '\n':
                if not was_carriage_return:
                    # NOTE: carriage return incremented line number,
                    #       but we reset the column index to zero
                    column_index = 0
                was_carriage_return = False
            elif unicode_scalar == '\r':
                line_index += 1
                column_index = 0
                was_carriage_return = True
            else:
                column_index += 1
                was_carriage_return = False

            # prepare for next iteration
            byte_offset += len(unicode_scalar.encode(input_encoding))

    # check validity of result
    if begin is None:
        raise ValueError(f"Byte range {byte_range[0]}–{byte_range[1]} is outside content of {input_encoding} text file '{filepath}'")

    return begin, end, substring.replace('\r\n', '\n').replace('\r', '\n')


HIGHLIGHT = ('\033[0;34;49m', '\033[0m')
EMPHASIZE = ('\033[0;33;49m', '\033[0m')


def iterate_lines(content: bytes = None, content_encoding = 'utf-8', binary_width: int = 16) -> typing.Generator[typing.Tuple[int, int, bytes], None, None]:
    """Returns a generator over the lines of a file based on `content_encoding`.
    If the file has ``content_encoding == 'binary'``, a line always contains exactly `binary_width` bytes.
    The generator returns the absolute position of the first character of the line, the absolute position of
    the line break in this line, and also the content of the line including its potential line break
    (potential, because the last line has none).
    """
    if content_encoding.lower() not in {'utf-8', 'ascii', 'binary', 'bin', ''}:
        raise ValueError("Content encoding must be one of {utf-8, ascii, binary}, but got '" + str(content_encoding) + '"')

    content_encoding = content_encoding.lower()
    if content_encoding == 'ascii':
        pos = 0
        while True:
            next_pos = min(content.find('\n', pos), content.find('\r', pos))
            if next_pos == -1:
                yield pos, len(content), content[pos:]
                return

            without_linebreak = next_pos
            if content[next_pos] == '\r' and len(content) > next_pos + 1 and content[next_pos + 1] == '\n':
                next_pos += 2
            else:
                next_pos += 1

            yield pos, without_linebreak, content[pos:next_pos]
            pos = next_pos

    elif content_encoding == 'utf-8':
        # via https://www.unicode.org/reports/tr14/ on 2023-10-30
        #   after BK, LF, and NL {000B 000C 2028 2029 000A 0085}
        #   after CR except between CR and LF {000D}
        pos = 0
        while True:
            # evaluate next positions
            cr_pos = content.find(b'\u000D', pos)
            if cr_pos == -1:
                cr_pos = float('inf')

            individual_pos = float('inf')
            for scalars in [c.encode('utf-8') for c in '\u000B\u000C\u2028\u2029\u000A\u0085']:
                n = content.find(scalars, pos)
                if n != -1 and n < individual_pos:
                    individual_pos = n

            # evaluate next line break
            if individual_pos < cr_pos:
                yield pos, individual_pos, content[pos:individual_pos + 1]
                pos = individual_pos + 1
            
            elif cr_pos < individual_pos:
                length = 2 if len(content) > cr_pos + 1 and content[cr_pos + 1] == '\n' else 2
                yield pos, cr_pos, content[pos:cr_pos + length]
                pos = cr_pos + length

            else:
                # both are infinite
                yield pos, len(content), content[pos:]
                return

    else:
        for offset in range(0, binary_width, binary_width):
            subcontent = content[offset:offset + binary_width]
            yield offset, offset + binary_width, subcontent


def utf8_lines_before_until_after_byte_range(byte_range: typing.Tuple[int, int], content: bytes = None) -> typing.List[typing.Tuple[int, int, bytes]]:
    """Fetch all lines of this UTF-8 file. Assume `byte_range` points to line index 5,
    then we return lines of indices 4, 5, and 6 (i.e. indices covering `byte_range` as well as -1 and +1).
    The list contains tuples with the line index, the absolute byte offset of the first character of the line,
    and the content of the line without a potential final line break character.
    """
    all_lines = list(iterate_lines(content, 'utf-8'))

    # determine index of first line
    first_line_index = -1
    for i, (start_offset, break_offset, line) in enumerate(all_lines):
        if start_offset <= byte_range[0] <= start_offset + len(line):
            first_line_index = i
            break

    # determine index of last line
    last_line_index = -1
    for i, (start_offset, break_offset, line) in enumerate(all_lines):
        if start_offset <= byte_range[1] <= start_offset + len(line):
            last_line_index = i
            break

    # exception if those are not found
    if first_line_index == -1:
        raise ValueError(f"Could not identify line containing byte offset {first_line_index} - does it occur in a content of length {len(content)}?")
    elif last_line_index == -1:
        raise ValueError(f"Could not identify line containing byte offset {last_line_index} - does it occur in a content of length {len(content)}?")

    for line_index in range(max(first_line_index - 1, 0), min(last_line_index + 1, len(all_lines) - 1) + 1):
        line_data = all_lines[line_index]
        yield line_index, line_data[0], line_data[2][0:line_data[1] - line_data[0]]


class Line:
    def __init__(self, byte_offset: int, line_number: int = None, content: bytes = None, highlight: (typing.Optional[int], typing.Optional[int]) = (0, 0)):
        self.byte_offset = byte_offset  # this is currently not printed and just for documentation purposes
        self.line_number = line_number
        self.content = content
        self.highlight = highlight  # None means infinite index; to disable highlights, set highlight to (a, b) where b <= a

    def print_to_filedescriptor(self, indent=0, out_file=sys.stderr):
        # TODO should it simply raise an exception if it is not UTF-8 content?
        indentation = '  ' * indent
        lineno = ' {:>5}'.format(self.line_number) if (self.line_number and self.line_number > 0) else (' ' * 6)

        if self.highlight[0] is None and self.highlight[1] is None:
            content = HIGHLIGHT[0] + self.content.decode('utf-8') + HIGHLIGHT[1]
        elif self.highlight[0] is not None and self.highlight[1] is None:
            content = self.content[0:self.highlight[0]].decode('utf-8') + HIGHLIGHT[0] + self.content[self.highlight[0]:].decode('utf-8') + HIGHLIGHT[1]
        elif self.highlight[0] is None and self.highlight[1] is not None:
            content = HIGHLIGHT[0] + self.content[0:self.highlight[1]].decode('utf-8') + HIGHLIGHT[1] + self.content[self.highlight[1]:].decode('utf-8')
        elif self.highlight[0] is not None and self.highlight[1] is not None:
            if self.highlight[1] <= self.highlight[0]:
                content = self.content.decode('utf-8')
            else:
                content = (
                    self.content[0:self.highlight[0]].decode('utf-8') + HIGHLIGHT[0] +
                    self.content[self.highlight[0]:self.highlight[1]].decode('utf-8') + HIGHLIGHT[1] +
                    self.content[self.highlight[1]:].decode('utf-8')
                )

        print(f'{indentation}{lineno} | {content}', file=out_file)


class Frame:
    def __init__(self, description: str = None, lines: typing.List[Line] = None):
        assert description is None or len(description.splitlines()) < 2, repr(description)
        self.description = description
        self.lines = lines or []

    def print_to_filedescriptor(self, indent=0, out_file=sys.stderr):
        if self.description:
            print('{}{}→{} {}'.format('  ' * indent, EMPHASIZE[0], EMPHASIZE[1], self.description), file=out_file)
        for line in self.lines:
            line.print_to_filedescriptor(indent=indent, out_file=out_file)


class Resource:
    def __init__(self, source, additional_info: str = None):
        self.source = source
        self.additional_info = additional_info
        self.frames = []

    def add_frame(self, frame: Frame):
        self.frames.append(frame)

    def print_to_filedescriptor(self, indent=0, out_file=sys.stderr):
        additional = ' ({})'.format(self.additional_info) if self.additional_info else ''
        print(f"in '{EMPHASIZE[0]}{self.source}{EMPHASIZE[1]}'{additional}", file=out_file)
        for frame in self.frames:
            frame.print_to_filedescriptor(indent=indent, out_file=out_file)


class SourceFileErrorLevel(enum.Enum):
    INFO = 1
    WARN = 2
    ERROR = 3

    def represent(self):
        match self:
            case self.INFO:
                return '\033[0;37;49mINFO\033[0m'
            case self.WARN:
                return '\033[0;33;49mWARN\033[0m'
            case self.ERROR:
                return '\033[0;31;49mERROR\033[0m'


class SourceFileError:
    def __init__(self, message: str, level: SourceFileErrorLevel = SourceFileErrorLevel.ERROR, identifier: str = None, resources: typing.List[Resource] = []):
        assert len(message.splitlines()) < 2, repr(message)
        self.message = message
        self.level = level
        self.identifier = identifier
        self.resources = resources or []

    def add_resource(self, res: Resource):
        self.resources.append(res)

    @staticmethod
    def constructor_arguments_from_single_file(file_path: str, error_message: str, byte_ranges: typing.List[typing.Tuple[int, int]], error_identifier: str = None, file_info: str = '', level: SourceFileErrorLevel = SourceFileErrorLevel.ERROR) -> typing.Any:
        res = Resource(file_path, additional_info=file_info)

        with open(file_path, 'rb') as fd:
            content = fd.read()

            for byte_range in byte_ranges:
                lines = []
                for line_data in utf8_lines_before_until_after_byte_range(byte_range, content):
                    line_number = line_data[0] + 1
                    line_byte_offset = line_data[1]
                    line_content = line_data[2]

                    starts_in_line = line_byte_offset < byte_range[0] <= line_byte_offset + len(line_content)
                    ends_in_line = line_byte_offset < byte_range[1] <= line_byte_offset + len(line_content)
                    inside_highlight = byte_range[0] < line_byte_offset and line_byte_offset + len(line_content) < byte_range[1]

                    if starts_in_line and ends_in_line:
                        highlight = (byte_range[0] - line_byte_offset, byte_range[1] - line_byte_offset)
                    elif starts_in_line:
                        highlight = (byte_range[0] - line_byte_offset, None)
                    elif ends_in_line:
                        highlight = (None, byte_range[1] - line_byte_offset)
                    elif inside_highlight:
                        highlight = (None, None)
                    else:
                        highlight = (0, 0) # do not highlight anything

                    lines.append(Line(line_byte_offset, line_number, line_content, highlight))
                res.add_frame(Frame(None, lines))

        return (error_message, level, error_identifier, [res])

    @staticmethod
    def from_single_file(file_path: str, error_message: str, byte_ranges: typing.List[typing.Tuple[int, int]], error_identifier: str = None, file_info: str = '', level: SourceFileErrorLevel = SourceFileErrorLevel.ERROR, constructor=None) -> typing.Any:
        if constructor is None:
            constructor = SourceFileError
        return constructor(*SourceFileError.constructor_arguments_from_single_file(file_path, error_message, byte_ranges, error_identifier, file_info, level))

    def print_to_filedescriptor(self, indent=0, out_file=sys.stderr):
        for res in self.resources:
            res.print_to_filedescriptor(indent=indent, out_file=out_file)
            print('', file=out_file)

        identifier = '[{}]'.format(self.identifier) if self.identifier else ''
        print('{}{}: {}'.format(self.level.represent(), identifier, self.message), file=out_file)


class SpecificationError(SourceFileError):
    """An error in the HostVx specification file"""

    def __init__(self, message: str, level: SourceFileErrorLevel = SourceFileErrorLevel.ERROR, identifier: str = None, resources: typing.List[Resource] = []):
        super().__init__(message, level, identifier, resources)

    @staticmethod
    def in_file(source_file_path, error_message, child, parent_group):
        c = child.pos
        p = parent_group.pos

        args = SourceFileError.constructor_arguments_from_single_file(source_file_path, error_message, [p, c], file_info='specification file')
        return SpecificationError(*args)


class LexingError(SourceFileError):
    """An error while tokenizing the specification file"""

    def __init__(self, message: str, level: SourceFileErrorLevel = SourceFileErrorLevel.ERROR, identifier: str = None, resources: typing.List[Resource] = []):
        super().__init__(message, level, identifier, resources)

    @staticmethod
    def in_file(source_file_path, error_message, child, parent_group):
        c = child.pos
        p = parent_group.pos

        args = SourceFileError.constructor_arguments_from_single_file(source_file_path, error_message, [p, c], file_info='specification file')
        return LexingError(*args)



### ----------------- LISP lexer & parser ----------------------- ###

# raw characters of the syntax, I want to match
OPEN_PAREN = '('
CLOSE_PAREN = ')'
QUOTE_SINGLE = "'"
QUOTE_DOUBLE = '"'


class LexingState(enum.Enum):
    INIT = 1
    OPENED = 2
    CLOSED = 3
    READ_FUNC = 4
    READ_SEPARATOR = 5
    READ_STRING = 6
    READ_SINGLE_QUOTE = 7
    READ_DOUBLE_QUOTE = 8
    END_QUOTE = 9


class LexedToken(enum.Enum):
    OPEN_GROUP = 1  # start a new group
    STRING = 2  # gives a new string argument
    CLOSE_GROUP = 3  # terminates a group
    SKIP_WHITESPACE = 4  # whitespace not required to understand semantics
    SKIP_SINGLE_QUOTE = 5  # single quote not required to understand semantics
    SKIP_DOUBLE_QUOTE = 6  # double quote not required to understand semantics


class LispLexer:
    def __init__(self, source):
        self.state = LexingState.INIT
        self.scalar_id = 0
        self.column_id = 0
        self.line_id = 0
        self.start_string = None
        self.source = source

    def unexpected(self, char: str):
        assert len(char) == 1
        hex_id = 'U+{:04X}'.format(ord('A'))
        try:
            scalar_name = ' (' + unicodedata.name('A') + ')'
        except ValueError:
            scalar_name = ''
        context = {
            LexingState.INIT: 'while expecting the first group',
            LexingState.OPENED: 'after reading an opener of a new group',
            LexingState.CLOSED: 'after reading a group terminator',
            LexingState.READ_FUNC: 'while reading the function of a group',
            LexingState.READ_SEPARATOR: 'after reading a separator of arguments',
            LexingState.READ_STRING: 'while reading a string argument',
            LexingState.READ_SINGLE_QUOTE: 'when starting a single-quoted string',
            LexingState.READ_DOUBLE_QUOTE: 'when starting a double-quoted string',
            LexingState.END_QUOTE: 'when finishing a quoted string'
        }[self.state]

        raise LexingError.in_file(self.source, f"Unexpected character '{char}' ({hex_id}{scalar_name}) at index {self.scalar_id} (line {self.line_id + 1} column {self.column_id + 1}) {context}")

    def state_init(self, char: str):
        if char.isspace():
            return LexingState.INIT, [(LexedToken.SKIP_WHITESPACE, self.scalar_id, self.scalar_id + 1)]
        elif char == OPEN_PAREN:
            return LexingState.OPENED, [(LexedToken.OPEN_GROUP, self.scalar_id, self.scalar_id + 1)]
        else:
            self.unexpected(char)

    def state_opened(self, char: str):
        ...




    def consume(self, char: str):
        if self.state == LexingState.INIT:
            self.state, result = self.state_init(char)
            yield from result
        elif self.state == LexingState.OPENED:
            self.state, result = self.state_opened(char)
            yield from result
        elif self.state == LexingState.CLOSED:
            self.state, result = self.state_closed(char)
            yield from result
        elif self.state == LexingState.READ_FUNC:
            self.state, result = self.state_read_func(char)
            yield from result
        elif self.state == LexingState.READ_SEPARATOR:
            self.state, result = self.state_read_separator(char)
            yield from result
        elif self.state == LexingState.READ_STRING:
            self.state, result = self.state_read_string(char)
            yield from result
        elif self.state == LexingState.READ_SINGLE_QUOTE:
            self.state, result = self.state_read_single_quote(char)
            yield from result
        elif self.state == LexingState.READ_DOUBLE_QUOTE:
            self.state, result = self.state_read_double_quote(char)
            yield from result
        elif self.state == LexingState.END_QUOTE:
            self.state, result = self.state_end_quote(char)
            yield from result
        else:
            raise ValueError(f"Unknown state '{self.state}'")

    def feed(self, next_characters: str):
        for char in next_characters:
            yield from self.consume(char)
            self.scalar_id += 1
            if char == '\n':
                self.column_id = 0
                self.line_id += 1
            else:
                self.column_id += 1


### --------------------------------- parser ------------------------------- ###

class LispParser:
    """`LispParser` consuming the tokens emitted by `LispLexer`"""

    def __init__(self, source_path, source_text):
        self.source_path = source_path
        self.source = source_text
        self.tree = Group()
        self.tree.add_argument(String('root', (None, None)))
        self.current_path = []

    def consume_token(self, token):
        tok, start, end = token
        parent = self.tree
        for component in self.current_path:
            parent = parent.arguments[component]

        match tok:
            case LexedToken.OPEN_GROUP:
                new_group = Group()
                new_group.set_start_position(start)
                new_group.set_max_end_position(end)
                child_index = parent.add_argument(new_group)
                self.current_path.append(child_index)
            case LexedToken.STRING:
                parent.add_argument(String(self.source[start:end], (start, end)))
                parent.set_max_end_position(end)
            case LexedToken.CLOSE_GROUP:
                parent.set_max_end_position(end)
                self.current_path.pop()
            case _:
                pass

    def consume_tokens(self, next_tokens):
        for token in next_tokens:
            self.consume_token(token)

    def finalize(self, final_lex_state, final_lex_start_scalar_id=None):
        if self.current_path:
            pos_info = ''
            if final_lex_start_scalar_id:
                token_start, _, suffix = byte_range_to_position_information(self.source_path, (final_lex_start_scalar_id, final_lex_start_scalar_id + 1))
                pos_info = f' at line {token_start[1]} column {token_start[2]}'
            raise ValueError(f'Not all open groups have been closed, terminating in state {final_lex_state}' + pos_info)
        return self.tree


class String:
    """`String` is a pure string occuring in the implemented LISP dialect"""

    def __init__(self, text, pos):
        self.text = text
        self.pos = pos
        self.is_group = False

    def __str__(self):
        if '"' not in self.text and "'" not in self.text:
            return self.text
        elif '"' not in self.text:
            return f"'{self.text}'"
        elif "'" not in self.text:
            return f'"{self.text}"'
        else:
            escaped = self.text.replace("'", "' \"'\" '")
            return f"(str '{escaped}')"

    def __repr__(self):
        return repr(self.text)


class Group:
    """`Group` is the list-like type occuring in the LISP dialect"""

    def __init__(self):
        self.pos = (None, None)
        self.arguments = []
        self.is_group = True

    @property
    def function(self):
        return self.arguments[0].text

    @property
    def args(self):
        return self.arguments[1:]

    def set_start_position(self, start):
        self.pos = (start, self.pos[1])

    def set_max_end_position(self, end):
        if end is None:
            raise ValueError("Argument 'end' must not be None")
        self.pos = (self.pos[0], end if self.pos[1] is None else max(self.pos[1], end))

    def add_argument(self, arg):
        # NOTE: first argument must not be an instance of `Group`
        if not self.arguments and arg.is_group:
            print(arg, self.pos)
            raise ValueError("First argument of a Group must not be a Group itself")
        self.arguments.append(arg)
        return len(self.arguments) - 1

    def __str__(self):
        # non-recursive representation
        return '({})'.format(' '.join('…' if arg.is_group else str(arg) for arg in self.arguments))

    def __repr__(self):
        # recursive representation
        return "[bytes {}–{}]({})".format(self.pos[0], self.pos[1], ' '.join(repr(arg) for arg in self.arguments))


### ---------------------------------- main -------------------------------- ###

def main(spec_file):
    """Read the specification file and represent it in an arbitrary way
    so the user can verify that the file is interpreted appropriately.
    """
    lex = LispLexer(spec_file)

    with open(spec_file) as fd:
        content = fd.read()
        par = LispParser(spec_file, content)

        for token in lex.feed(content):
            tok, start, end = token
            print(tok, repr(content[start:end]))

            par.consume_token(token)

        spec_tree = par.finalize()

    print()
    print(repr(spec_tree))



if __name__ == '__main__':
    loglevels = 'CRITICAL;DEBUG;ERROR;FATAL;INFO;NOTSET;WARN;WARNING'.split(';')

    parser = argparse.ArgumentParser(description=__doc__.strip())
    parser.add_argument('src', help='lispygraz file to lex')
    parser.add_argument('--log-level', dest='loglevel', default='DEBUG', choices=loglevels, help='loglevel for the logging module')
    parser.add_argument('--log-format', dest='logformat', default='%(asctime)s,%(levelname)s: %(message)s', help='log message for the logging module')

    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.loglevel), format=args.logformat)
    LOGGER.setLevel(getattr(logging, args.loglevel))

    exitcode = main(args.src) or 0
    sys.exit(exitcode)
